/*---------------------------------------------------------------------
Copyright Frédéric Legrand (2012)
  
This software is a computer program whose purpose is to solve Poisson
 equation with a finite difference method. A Gauss-Seidel relaxation 
method is used, with both CPU and GPU (OpenCL) implementation. 

This software is governed by the CeCILL  license under French law and
abiding by the rules of distribution of free software.  You can  use,  
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info".  

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 
 
The fact that you are presently reading this means that you have had 
knowledge of the CeCILL license and that you accept its terms.
 
-------------------------------------------------------------------------*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "poisson.h"
#ifdef OPENCL
#include <CL/cl.h>
#endif
#include <pthread.h>


char px,py; // puissance de 2 de la taille du domaine
short width,height; // taille du domaine
short largeur,hauteur; // taille du domaine + les bords
int taille,size;

// matrice creuse A 
float *G; // gauche
float *D; // droite
float *H; // haut
float *B; // bas
float *C; // centre

float *U; // valeurs à déterminer, solutions de AU=F
float *U1; // tableau pour les sorties
float *F; // terme source


// OpenCL
#ifdef OPENCL
cl_uint numPlatforms;
cl_device_type deviceType;
cl_platform_id platform;
cl_device_id device;
size_t maxWorkGroupSize;
cl_uint maxWorkItemDimensions;
size_t maxWorkItemSizes[3];
cl_context context;
cl_program OpenCLProgram;
cl_command_queue commandQueue;
cl_kernel iterationPaire, iterationImpaire, iterationJacobiOne, iterationJacobiTwo, arrayNorm, iterationPaireColonne, iterationImpaireColonne, iterationPaireLigne, iterationImpaireLigne;
cl_mem gpu_G,gpu_D,gpu_H,gpu_B,gpu_C,gpu_F,gpu_U,gpu_V;
#endif

//pthread
pthread_barrier_t threadBarrier;
typedef struct {
    int niter;
    int jStart;
    int jEnd;
    float w1;
    float w;
} threadArg;

#ifdef OPENCL
// programme openCL
const char* programSource[] = {
"__kernel void iterationPaire(__global float* D, __global float*G, __global float *H, __global float *B, __global float *C, __global float *F, __global float *U, short largeur, float w1, float w) {",
    "unsigned short i = get_global_id(0);",
    "unsigned short j = get_global_id(1);",
    "unsigned int c = i+j*largeur;",
    "if ((((i&1)+(j&1))&1)==0) { // bit de parité de i+j\n",
        "U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];",
    "}",
"}",
"__kernel void iterationImpaire(__global float* D, __global float*G, __global float *H, __global float *B, __global float *C, __global float *F, __global float *U, short largeur, float w1, float w) {",
    "unsigned short i = get_global_id(0);",
    "unsigned short j = get_global_id(1);",
    "unsigned int c = i+j*largeur;",
    "if ((((i&1)+(j&1))&1)==1) {",
        "U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];",
    "}",
"}",
"__kernel void iterationImpaireColonne(__global float* D, __global float*G, __global float *H, __global float *B, __global float *C, __global float *F, __global float *U, short largeur, short ligne, short colonne, float w1, float w) {",
    "unsigned short i = get_global_id(0);",
    "unsigned int c = colonne+i*largeur;",
    "if ((i&1)==1) {",
        "U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];",
    "}",
"}",
"__kernel void iterationImpaireLigne(__global float* D, __global float*G, __global float *H, __global float *B, __global float *C, __global float *F, __global float *U, short largeur, short ligne, short colonne, float w1, float w) {",
    "unsigned short i = get_global_id(0);",
    "unsigned int c = i+ligne*largeur;",
    "if ((i&1)==1) {",
        "U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];",
    "}",
"}",
"__kernel void iterationPaireColonne(__global float* D, __global float*G, __global float *H, __global float *B, __global float *C, __global float *F, __global float *U, short largeur, short ligne, short colonne, float w1, float w) {",
    "unsigned short i = get_global_id(0);",
    "unsigned int c = colonne+i*largeur;",
    "if ((i&1)==0) {",
        "U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];",
    "}",
    "if (i==1) {",
        "c = colonne+ligne*largeur;",
        "U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];",
    "}",
"}",
"__kernel void iterationPaireLigne(__global float* D, __global float*G, __global float *H, __global float *B, __global float *C, __global float *F, __global float *U, short largeur, short ligne, short colonne,float w1, float w) {",
    "unsigned short i = get_global_id(0);",
    "unsigned int c = i+ligne*largeur;",
    "if ((i&1)==0) {",
        "U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];",
    "}",
"}",
"__kernel void arrayNorm(__global float *U, __local float *buf, __global float *norm) {",
    "unsigned int i = get_local_id(0);",
    "unsigned int c = get_global_id(0);",
    "float u = U[c];",
    "buf[i] = u*u;",
    "barrier(CLK_LOCAL_MEM_FENCE);",
    "unsigned int j = get_local_size(0)/2;",
    "while (j!=0) {",
        "if (i < j) buf[i] += buf[i+j];",
        "barrier(CLK_LOCAL_MEM_FENCE);",
        "j /= 2;",
    "}",
    "if (i==0) norm[get_group_id(0)] = buf[0];",
"}"
};
#define SOURCE_NB_LINES 62
 
#define OPENCL_ERROR(e,msg) if (e<0) {printf("Erreur openCL : %s : %d\n",msg,e);return e;}

#endif


int poisson_open(int pw, int ph) {
    #ifdef OPENCL
    platform = 0; 
    #endif
    width = (short)pow(2,pw)+1;
    height = (short)pow(2,ph)+1;
    size = width*height;
    largeur = width + 2; hauteur = height + 2;
    taille = largeur*hauteur;
    G = (float*)malloc(sizeof(float)*taille);
    D = (float*)malloc(sizeof(float)*taille);
    H = (float*)malloc(sizeof(float)*taille);
    B = (float*)malloc(sizeof(float)*taille);
    C = (float*)malloc(sizeof(float)*taille);
    U = (float*)malloc(sizeof(float)*taille);
    F = (float*)malloc(sizeof(float)*taille);
    U1 = (float*)malloc(sizeof(float)*width*(2*height-1));
    return 0;
}

int poisson_close() {
    free(G); free(D); free(H); free(B); free(C); free(F); free(U);
    free(U1); 
    return 0;
}

int poisson_laplacien(float dx, float dy) {
    int k;
    float dx2,dy2;
    float r;
    dx2=dx*dx;
    dy2=dy*dy;
    r = dy2/dx2;
    for (k=0; k<taille; k++) {
        D[k] = G[k] = r;
        H[k] = B[k] = 1.0;
        C[k] = -2*(1+r);
        F[k] = 0.0;
        U[k] = 0.0;
    }
    return 0;
}

int poisson_laplacien_axial(float dz, float dr) {
    int i,j,k;
    float r;
    r = dr*dr/(dz*dz);
    for (i=0; i<width; i++) {
        j=0;
        k = (i+1)+largeur*(j+1);
        D[k] = G[k] = r/8.0;
        H[k] = 0.5;
        B[k] = 0.0;
        C[k] = -r/4.0-0.5;
        F[k] = 0.0;
        U[k] = 0.0;
        for (j=1; j<height; j++) {
            k = (i+1)+largeur*(j+1);
            D[k] = G[k] = j*r;
            H[k] = j+0.5;
            B[k] = j-0.5;
            C[k] = -2*j*(1.0+r);
            F[k] = 0.0;
            U[k] = 0.0;
        }
        
    }
    return 0;
}

int poisson_laplacien_axial_vect(float dz, float dr) {
    int i,j,k;
    float r;
    r = dr*dr/(dz*dz);
    for (i=0; i<width; i++) {
        j=0;
        k = (i+1)+largeur*(j+1);
        D[k] = G[k] = H[k] = B[k] = 0;
        C[k] = 1.0;
        U[k] = F[k] = 0;
        for (j=1; j<height; j++) {
            k = (i+1)+largeur*(j+1);
            D[k] = G[k] = r;
            H[k] = (j+1.0)/(j+0.5);
            B[k] = (j-1.0)/(j-0.5);
            C[k] = -(j/(j+0.5)+j/(j-0.5)+2*r);
            F[k] = 0.0;
            U[k] = 0.0;
        }
    }
    return 0;
}

int poisson_dirichlet_borders(float value) {
     int i,j,k;
     for (i=1; i<=width; i++) { // dirichlet sur les bords hauts et bas
         k = i+largeur;
         D[k] = G[k] = H[k] = B[k] = 0;
         C[k] = 1;
         U[k] = F[k] = value;
         k = i+(hauteur-2)*largeur;
         D[k] = G[k] = H[k] = B[k] = 0;
         C[k] = 1;
         U[k] = F[k] = value;
     }
     for (j=1; j<=height; j++) { // dirichlet sur les bords gauche et droite
         k = 1+j*largeur;
         D[k] = G[k] = H[k] = B[k] = 0;
         C[k] = 1;
         U[k] = F[k] = value;
         k = largeur-2+j*largeur;
         D[k] = G[k] = H[k] = B[k] = 0;
         C[k] = 1;
         U[k] = F[k] =value;
     }
     return 0;
}

int poisson_dirichlet_borders_axial(float value) {
     int i,j,k;
     for (i=1; i<=width; i++) { // dirichlet sur le bord haut
         k = i+(hauteur-2)*largeur;
         D[k] = G[k] = H[k] = B[k] = 0;
         C[k] = 1;
         U[k] = F[k] = value;
         
     }
     for (j=1; j<=height; j++) { // dirichlet sur les bords gauche et droite
         k = 1+j*largeur;
         D[k] = G[k] = H[k] = B[k] = 0;
         C[k] = 1;
         U[k] = F[k] = value;
         k = largeur-2+j*largeur;
         D[k] = G[k] = H[k] = B[k] = 0;
         C[k] = 1;
         U[k] = F[k] =value;
     }
     return 0;
}

int poisson_dirichlet_borders_axial_vect(float value) {
     int i,j,k;
     for (i=1; i<=width; i++) { // dirichlet sur le bord haut
         k = i+(hauteur-2)*largeur;
         D[k] = G[k] = H[k] = B[k] = 0;
         C[k] = 1;
         U[k] = F[k] = value;
         
     }
     for (j=1; j<=height; j++) { // dirichlet sur les bords gauche et droite
         k = 1+j*largeur;
         D[k] = G[k] = H[k] = B[k] = 0;
         C[k] = 1;
         U[k] = F[k] = value;
         k = largeur-2+j*largeur;
         D[k] = G[k] = H[k] = B[k] = 0;
         C[k] = 1;
         U[k] = F[k] =value;
     }
     return 0;
}
 
int poisson_dirichlet(int i, int j, float value) {
    int k;
    if ((i<0)||(j<0)||(i>=width)||(j>=height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (i+1)+largeur*(j+1);
    D[k] = G[k] = H[k] = B[k] = 0;
    C[k] = 1;
    U[k] = F[k] = value;
    return 0;
}

int poisson_source(int i, int j, float s, float dx, float dy) {
    int k;
    if ((i<0)||(j<0)||(i>=width)||(j>=height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (i+1)+largeur*(j+1);
    F[k] = dy*dy*s;
    return 0;
}

int poisson_source_axial(int i, int j, float s, float dz, float dr) {
    int k;
    if ((i<0)||(j<0)||(i>=width)||(j>=height)) return POISSON_INDEX_OUT_OF_RANGE;
    k = (i+1)+largeur*(j+1);
    if (j==0) {
        F[k] = dr*dr/8.0;   
    }
    else {
        F[k] = j*dr*dr*s;
    }
    return 0;
}

int poisson_source_polygon(intArray ilist, intArray jlist, float s, float dx, float dy) {
    int k;
    if ((ilist.ndims!=1)||(jlist.ndims!=1)) return POISSON_BAD_ARRAY_DIM;
    if (ilist.dims[0]!=jlist.dims[0]) return POISSON_BAD_ARRAY_DIM;
    for (k=0; k<ilist.dims[0]; k++) poisson_source(ilist.data[k],jlist.data[k],s,dx,dy);
    return 0;   
}

int poisson_source_polygon_axial(intArray ilist, intArray jlist, float s, float dz, float dr) {
    int k;
    if ((ilist.ndims!=1)||(jlist.ndims!=1)) return POISSON_BAD_ARRAY_DIM;
    if (ilist.dims[0]!=jlist.dims[0]) return POISSON_BAD_ARRAY_DIM;
    for (k=0; k<ilist.dims[0]; k++) poisson_source_axial(ilist.data[k],jlist.data[k],s,dz,dr);
    return 0;   
}

int poisson_dirichlet_polygon(intArray ilist, intArray jlist,float value) {
    int k;
    if ((ilist.ndims!=1)||(jlist.ndims!=1)) return POISSON_BAD_ARRAY_DIM;
    if (ilist.dims[0]!=jlist.dims[0]) return POISSON_BAD_ARRAY_DIM;
    for (k=0; k<ilist.dims[0]; k++) poisson_dirichlet(ilist.data[k],jlist.data[k],value);
    return 0;
}

int poisson_neumann(int ip, int jp, int i, int j, int in, int jn, float source, float derivX, float derivY, float dx, float dy) {
    //ip,jp : previous point
    //i,j : point
    //in,jn : next point
    int di1,di2,dj1,dj2;
    int k;
    float dx2,dy2,r;
    dx2=dx*dx;
    dy2=dy*dy;
    r=dy2/dx2;
    di1 = i-ip;
    di2 = in-i;
    dj1 = j-jp;
    dj2 = jn-j;
    k = (i+1)+largeur*(j+1);
    if ((abs(di1+dj1)!=1)||(abs(di2+dj2)!=1)) return POISSON_BAD_INDEX_SPEC;
    if ((di1==di2)&&(dj1==dj2)) { // segment rectiligne
        if (di1==1) {
            
            D[k] = G[k] = 0.5*r; H[k] = 1.0; B[k] = 0.0; C[k] = -1.0-r; F[k] = dy*derivY+0.5*dy2*source;
        }
        else if (di1==-1) {
            
            D[k] = G[k] = 0.5*r; H[k] = 0.0; B[k] = 1.0; C[k] = -1.0-r; F[k] = -dy*derivY+0.5*dy2*source;   
        }
        else if (dj1==1) {
            
            H[k] = B[k] = 0.5; G[k] = r; D[k] = 0.0; C[k] =-1.0-r; F[k] = -dy2/dx*derivX+0.5*dy2*source;
        }
        else if (dj1==-1) {
            
            H[k] = B[k] = 0.5; G[k] = 0.0; D[k] = r; C[k] =-1.0-r; F[k] = dy2/dx*derivX+0.5*dy2*source;    
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    else { // angle
        if ((di1==1)&&(dj2==1)) {
            G[k] = 0.5*r; D[k] =0.0; H[k] = 0.5; B[k] = 0.0; C[k] = -0.5*(1.0+r); F[k] = -0.5*dy2/dx*derivX+0.5*dy*derivY+0.25*dy2*source;  
        }
        else if ((di1==-1)&&(dj2==-1)) {
            G[k] = 0.0; D[k] = 0.5*r; H[k] = 0.0; B[k] = 0.5; C[k] = -0.5*(1.0+r); F[k] = 0.5*dy2/dx*derivX-0.5*dy*derivY+0.25*dy2*source;   
        }
        else if ((di2==1)&&(dj1==-1)) {
            G[k] = 0.0; D[k] = 0.5*r; H[k] = 0.5; B[k] = 0.0; C[k] = -0.5*(1.0+r); F[k] = 0.5*dy2/dx*derivX+0.5*dy*derivY+0.25*dy2*source;   
        }
        else if ((di2==-1)&&(dj1==1)) {
            G[k] = 0.5*r; D[k] = 0.0; H[k] = 0.0; B[k] = 0.5; C[k] = -0.5*(1.0+r); F[k] = -0.5*dy2/dx*derivX-0.5*dy*derivY+0.25*dy2*source;  
        }
        else if ((di1==1)&&(dj2==-1)) {
            G[k] = 0.5*r; D[k] = r; H[k] = 1.0; B[k] = 0.5; C[k] = -1.5*(1.0+r); F[k] = 0.5*dy2/dx*derivX+0.5*dy*derivY+0.75*dy2*source;
        }
        else if ((di1==-1)&&(dj2==1)) {
            G[k] = r; D[k] = 0.5*r; H[k] = 0.5; B[k] = 1.0; C[k] = -1.5*(1.0+r); F[k] = -0.5*dy2/dx*derivX-0.5*dy*derivY+0.75*dy2*source;
        }
        else if ((dj1==-1)&&(di2==-1)) {
            G[k] = 0.5*r; D[k] = r; H[k] = 0.5; B[k] = 1.0; C[k] = -1.5*(1.0+r); F[k] = 0.5*dy2/dx*derivX-0.5*dy*derivY+0.75*dy2*source;
        }
        else if ((dj1==1)&&(di2==1)) {
            G[k] = r; D[k] = 0.5*r; H[k] = 1.0; B[k] = 0.5; C[k] = -1.5*(1.0+r); F[k] = -0.5*dy2/dx*derivX+0.5*dy*derivY+0.75*dy2*source;   
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    
    return 0;
}

int poisson_neumann_axial(int ip, int jp, int i, int j, int in, int jn, float source, float derivX, float derivY, float dx, float dy) {
    //ip,jp : previous point
    //i,j : point
    //in,jn : next point
    int di1,di2,dj1,dj2;
    int k;
    float dx2,dy2,r;
    dx2=dx*dx;
    dy2=dy*dy;
    r=dy2/dx2;
    di1 = i-ip;
    di2 = in-i;
    dj1 = j-jp;
    dj2 = jn-j;
    k = (i+1)+largeur*(j+1);
    if ((abs(di1+dj1)!=1)||(abs(di2+dj2)!=1)) return POISSON_BAD_INDEX_SPEC;
    if ((di1==di2)&&(dj1==dj2)) { // segment rectiligne
        if (di1==1) {
            G[k]=D[k]=0.5*(j+0.25)*r; B[k] = 0.0; H[k] = (j+0.5); C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = j*dy*derivY+0.5*(j+0.25)*dy2*source;
            
        }
        else if (di1==-1) {
            G[k]=D[k]=0.5*(j-0.25)*r; B[k] = j-0.5; H[k] = 0.0; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = -j*dy*derivY+0.5*(j-0.25)*dy2*source;
            
        }
        else if (dj1==1) {
            G[k] = j*r; D[k] = 0.0; H[k] = 0.5*(j+0.5); B[k] = 0.5*(j-0.5); C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = -j*dy2/dx*derivX+0.5*j*dy2*source;
        }
        else if (dj1==-1) {
            G[k] = 0.0; D[k] = j*r; H[k] = 0.5*(j+0.5); B[k] = 0.5*(j-0.5); C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = j*dy2/dx*derivX+0.5*j*dy2*source;
                
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    else { // angle
        if ((di1==1)&&(dj2==1)) {
              G[k] = 0.5*(j+0.25)*r; D[k] = 0.0; H[k] = 0.5*(j+0.5); B[k] = 0.0; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = -0.5*(j+0.25)*dy2/dx*derivX+0.5*j*dy*derivY+0.25*(j+0.25)*dy2*source;
              
        }
        else if ((di1==-1)&&(dj2==-1)) {
              G[k] = 0.0; D[k] = 0.5*(j-0.25)*r; H[k] = 0.0; B[k] = 0.5*(j-0.5); C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = 0.5*(j-0.25)*dy2/dx*derivX-0.5*j*dy*derivY+0.25*(j-0.25)*dy2*source;
              
        }
        else if ((di2==1)&&(dj1==-1)) {
               G[0] = 0.0; D[k] = 0.5*(j+0.25)*r; H[k] = 0.5*(j+0.5); B[k] = 0.0;  C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = 0.5*(j+0.25)*dy2/dx*derivX-0.5*j*dy*derivY+0.25*(j+0.25)*dy2*source;
        }
        else if ((di2==-1)&&(dj1==1)) {
               G[k] = 0.5*(j-0.25)*r; D[k] = 0.0; H[k] = 0.0; B[k] = 0.5*(j-0.5); C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = -0.5*(j-0.25)*dy2/dx*derivX-0.5*j*dy*derivY+0.25*(j-0.25)*dy2*source;
        }
        else if ((di1==1)&&(dj2==-1)) {
                G[k] = 0.5*(j+0.25)*r; D[k] = j*r; H[k] = j+0.5; B[k] = 0.5*(j-0.5); C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = -0.5*(j+0.25)*dy2/dx*derivX-0.5*j*dy*derivY+(0.5*(j-0.25)+0.25*(j+0.25))*dy2*source;
        }
        else if ((di1==-1)&&(dj2==1)) {
                G[k] = j*r; D[k] = 0.5*(j-0.25)*r; H[k] = 0.5*(j+0.5); B[k] = (j-0.5); C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = 0.5*(j-0.25)*dy2/dx*derivX+0.5*j*dy*derivY+(0.5*(j+0.25)+0.25*(j-0.25))*dy2*source;
        }
        else if ((dj1==-1)&&(di2==-1)) {
                G[k] = 0.5*(j-0.25)*r; D[k] = j*r; H[k] = 0.5*(j+0.5); B[k] = j-0.5; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] =  0.5*(j+0.25)*dy2/dx*derivX-0.5*j*dy*derivY+(0.5*(j-0.25)+0.25*(j+0.25))*dy2*source;
        }
        else if ((dj1==1)&&(di2==1)) {
               G[k] = j*r; D[k] = 0.5*(j+0.25)*r; H[k] = j+0.5; B[k] = 0.5*(j-0.5); C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] =  -0.5*(j-0.25)*dy2/dx*derivX+0.5*j*dy*derivY+(0.5*(j+0.25)+0.25*(j-0.25))*dy2*source;
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    
    return 0;
}

int poisson_neumann_axial_vect(int ip, int jp, int i, int j, int in, int jn, float source, float derivX, float derivY, float dx, float dy) {
    //ip,jp : previous point
    //i,j : point
    //in,jn : next point
    int di1,di2,dj1,dj2;
    int k;
    float dx2,dy2,r;
    dx2=dx*dx;
    dy2=dy*dy;
    r=dy2/dx2;
    di1 = i-ip;
    di2 = in-i;
    dj1 = j-jp;
    dj2 = jn-j;
    k = (i+1)+largeur*(j+1);
    if ((abs(di1+dj1)!=1)||(abs(di2+dj2)!=1)) return POISSON_BAD_INDEX_SPEC;
    if ((di1==di2)&&(dj1==dj2)) { // segment rectiligne
        if (di1==1) {
            
            D[k] = G[k] = 0.5*r; H[k] = (j+1.0)/(j+0.5); B[k] = 0.0; C[k] = -(j/(j+0.5)+r); F[k] = dy*derivY+0.5*dy2*source;
        }
        else if (di1==-1) {
            
            D[k] = G[k] = 0.5*r; H[k] = 0.0; B[k] = (j-1.0)/(j-0.5); C[k] = -(j/(j-0.5)+r); F[k] = -dy*derivY+0.5*dy2*source;   
        }
        else if (dj1==1) {
            
            G[k] = r; D[k] = 0.0; H[k] = 0.5*(j+1)/(j+0.5); B[k] = 0.5*(j-1)/(j-0.5); C[k] =-(0.5*j/(j+0.5)+0.5*j/(j-0.5)+r); F[k] = -dy2/dx*derivX+0.5*dy2*source;
        }
        else if (dj1==-1) {
            
            G[k] = 0.0; D[k] = r; H[k] = 0.5*(j+1)/(j+0.5); B[k] = 0.5*(j-1)/(j-0.5); C[k] =-(0.5*j/(j+0.5)+0.5*j/(j-0.5)+r); F[k] = dy2/dx*derivX+0.5*dy2*source;    
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    else { // angle
        if ((di1==1)&&(dj2==1)) {
            G[k] = 0.5*r; D[k] =0.0; H[k] = 0.5*(j+1.0)/(j+0.5); B[k] = 0.0; C[k] = -(0.5*j/(j+0.5)+r); F[k] = -0.5*dy2/dx*derivX+0.5*dy*derivY+0.25*dy2*source;  
        }
        else if ((di1==-1)&&(dj2==-1)) {
            G[k] = 0.0; D[k] = 0.5*r; H[k] = 0.0; B[k] = 0.5*(j-1.0)/(j-0.5); C[k] = -(0.5*j/(j-0.5)+r); F[k] = 0.5*dy2/dx*derivX-0.5*dy*derivY+0.25*dy2*source;   
        }
        else if ((di2==1)&&(dj1==-1)) {
            G[k] = 0.0; D[k] = 0.5*r; H[k] = 0.5*(j+1.0)/(j+0.5); B[k] = 0.0; C[k] = -(0.5*j/(j+0.5)+r); F[k] = 0.5*dy2/dx*derivX+0.5*dy*derivY+0.25*dy2*source;   
        }
        else if ((di2==-1)&&(dj1==1)) {
            G[k] = 0.5*r; D[k] = 0.0; H[k] = 0.0; B[k] = 0.5*(j-1.0)/(j-0.5); C[k] = -(0.5*j/(j-0.5)+r); F[k] = -0.5*dy2/dx*derivX-0.5*dy*derivY+0.25*dy2*source;  
        }
        else if ((di1==1)&&(dj2==-1)) {
            G[k] = 0.5*r; D[k] = r; H[k] = (j+1.0)/(j+0.5); B[k] = 0.5*(j-1.0)/(j-0.5); C[k] = -(j/(j+0.5)+0.5*j/(j-0.5)+1.5*r); F[k] = 0.5*dy2/dx*derivX+0.5*dy*derivY+0.75*dy2*source;
        }
        else if ((di1==-1)&&(dj2==1)) {
            G[k] = r; D[k] = 0.5*r; H[k] = 0.5*(j+1.0)/(j+0.5); B[k] = (j-1.0)/(j-0.5); C[k] = -(0.5*j/(j+0.5)+j/(j-0.5)+1.5*r); F[k] = -0.5*dy2/dx*derivX-0.5*dy*derivY+0.75*dy2*source;
        }
        else if ((dj1==-1)&&(di2==-1)) {
            G[k] = 0.5*r; D[k] = r; H[k] = 0.5*(j+1.0)/(j+0.5); B[k] = (j-1.0)/(j-0.5); C[k] = -(0.5*j/(j+0.5)+j/(j-0.5)+1.5*r); F[k] = 0.5*dy2/dx*derivX-0.5*dy*derivY+0.75*dy2*source;
        }
        else if ((dj1==1)&&(di2==1)) {
            G[k] = r; D[k] = 0.5*r; H[k] = (j+1.0)/(j+0.5); B[k] = 0.5*(j-1.0)/(j-0.5); C[k] = -(j/(j+0.5)+0.5*j/(j-0.5)+1.5*r); F[k] = -0.5*dy2/dx*derivX+0.5*dy*derivY+0.75*dy2*source;   
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    
    return 0;
}

int poisson_interface(int ip, int jp, int i, int j, int in, int jn, float source,float a1, float a2, float dx, float dy) { // interface entre deux milieux différents
    //ip,jp : previous point
    //i,j : point
    //in,jn : next point
    int di1,di2,dj1,dj2;
    int k;
    float dx2,dy2,r;
    dx2=dx*dx; 
    dy2=dy*dy;
    r=dy2/dx2;
    di1 = i-ip;
    di2 = in-i;
    dj1 = j-jp;
    dj2 = jn-j;
    k = (i+1)+largeur*(j+1);
    if ((abs(di1+dj1)!=1)||(abs(di2+dj2)!=1)) return POISSON_BAD_INDEX_SPEC;
    if ((di1==di2)&&(dj1==dj2)) { // segment rectiligne
        if (di1==1) {
            D[k] = G[k] =r*(a1+a2)/2; H[k] = a1; B[k] = a2; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = dy2*source;
        }
        else if (di1==-1) {
            D[k] = G[k] = r*(a1+a2)/2; H[k] = a2; B[k] = a1; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = dy2*source;   
        }
        else if (dj1==1) {
            H[k] = B[k] = (a1+a2)/2; G[k] = a1*r; D[k] = a2*r; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = dy2*source; 
        }
        else if (dj1==-1) {
            H[k] = B[k] = (a1+a2)/2; G[k] = a2*r; D[k] = a1*r; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = dy2*source;    
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    else { // angle
        if ((di1==1)&&(dj2==1)) {
            G[k] = r*(a1+a2)/2; D[k] = r*a2; B[k] = a2; H[k] = (a1+a2)/2; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = dy2*source; 
        }
        else if ((di1==-1)&&(dj2==-1)) {
            G[k] = r*a2; D[k] = r*(a1+a2)/2; B[k] = (a1+a2)/2; H[k] = a2; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = dy2*source;  
        }
        else if ((di2==1)&&(dj1==-1)) {
            G[k] = r*a2; D[k] = r*(a1+a2)/2; B[k] = a2; H[k] = (a1+a2)/2; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = dy2*source;   
        }
        else if ((di2==-1)&&(dj1==1)) {
             G[k] = r*(a1+a2)/2; D[k] = r*a2; B[k] = (a1+a2)/2; H[k] = a2; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = dy2*source; 
        }
        else if ((di1==1)&&(dj2==-1)) {
           G[k] = r*(a1+a2)/2; D[k] = r*a1; B[k] = (a1+a2)/2; H[k] = a1; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = dy2*source; 
        }
        else if ((di1==-1)&&(dj2==1)) {
           G[k] = r*a1; D[k] = r*(a1+a2)/2; B[k] = a1; H[k] = (a1+a2)/2; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = dy2*source; 
        }
        else if ((dj1==-1)&&(di2==-1)) {
            G[k] = r*(a1+a2)/2; D[k] = r*a1; B[k] = a1; H[k] = (a1+a2)/2; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = dy2*source; 
        }
        else if ((dj1==1)&&(di2==1)) {
             G[k] = r*a1; D[k] = r*(a1+a2)/2; B[k] = (a1+a2)/2; H[k] = a1; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = dy2*source;  
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    
    return 0;       
}

int poisson_interface_axial(int ip, int jp, int i, int j, int in, int jn, float source,float a1, float a2, float dx, float dy) { // interface entre deux milieux différents
    //ip,jp : previous point
    //i,j : point
    //in,jn : next point
    int di1,di2,dj1,dj2;
    int k;
    float dx2,dy2,r;
    dx2=dx*dx;
    dy2=dy*dy;
    r=dy2/dx2;
    di1 = i-ip;
    di2 = in-i;
    dj1 = j-jp;
    dj2 = jn-j;
    k = (i+1)+largeur*(j+1);
    if ((abs(di1+dj1)!=1)||(abs(di2+dj2)!=1)) return POISSON_BAD_INDEX_SPEC;
    if ((di1==di2)&&(dj1==dj2)) { // segment rectiligne
        if (di1==1) {
            D[k] = G[k] =j*r*(a1+a2)/2; H[k] = (j+0.5)*a1; B[k] = (j-0.5)*a2; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = j*dy2*source;
        }
        else if (di1==-1) {
            D[k] = G[k] = j*r*(a1+a2)/2; H[k] = (j+0.5)*a2; B[k] = (j-0.5)*a1; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = j*dy2*source;   
        }
        else if (dj1==1) {
            H[k] = (j+0.5)*(a1+a2)/2; B[k] = (j-0.5)*(a1+a2)/2; G[k] = j*a1*r; D[k] = j*a2*r; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = j*dy2*source; 
        }
        else if (dj1==-1) {
            H[k] = (j+0.5)*(a1+a2)/2;  B[k] =(j-0.5)*(a1+a2)/2; G[k] = j*a2*r; D[k] = j*a1*r; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = j*dy2*source;    
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    else { // angle
        if ((di1==1)&&(dj2==1)) {
            G[k] = j*r*(a1+a2)/2; D[k] = j*r*a2; B[k] = (j-0.5)*a2; H[k] = (j+0.5)*(a1+a2)/2; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = j*dy2*source; 
        }
        else if ((di1==-1)&&(dj2==-1)) {
            G[k] = j*r*a2; D[k] = j*r*(a1+a2)/2; B[k] = (j-0.5)*(a1+a2)/2; H[k] = (j+0.5)*a2; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = j*dy2*source;  
        }
        else if ((di2==1)&&(dj1==-1)) {
            G[k] = j*r*a2; D[k] = j*r*(a1+a2)/2; B[k] = (j-0.5)*a2; H[k] = (j+0.5)*(a1+a2)/2; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = j*dy2*source;   
        }
        else if ((di2==-1)&&(dj1==1)) {
             G[k] = j*r*(a1+a2)/2; D[k] = j*r*a2; B[k] = (j-0.5)*(a1+a2)/2; H[k] = (j+0.5)*a2; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = j*dy2*source; 
        }
        else if ((di1==1)&&(dj2==-1)) {
           G[k] = j*r*(a1+a2)/2; D[k] = j*r*a1; B[k] = (j-0.5)*(a1+a2)/2; H[k] = (j+0.5)*a1; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = j*dy2*source; 
        }
        else if ((di1==-1)&&(dj2==1)) {
           G[k] = j*r*a1; D[k] = j*r*(a1+a2)/2; B[k] = (j-0.5)*a1; H[k] = (j+0.5)*(a1+a2)/2; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = j*dy2*source; 
        }
        else if ((dj1==-1)&&(di2==-1)) {
            G[k] = j*r*(a1+a2)/2; D[k] = j*r*a1; B[k] = (j-0.5)*a1; H[k] = (j+0.5)*(a1+a2)/2; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = j*dy2*source; 
        }
        else if ((dj1==1)&&(di2==1)) {
             G[k] = j*r*a1; D[k] = j*r*(a1+a2)/2; B[k] = (j-0.5)*(a1+a2)/2; H[k] = (j+0.5)*a1; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = j*dy2*source;  
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    
    return 0;       
}

int poisson_interface_axial_vect(int ip, int jp, int i, int j, int in, int jn, float source,float a1, float a2, float dx, float dy) { // interface entre deux milieux différents
    //ip,jp : previous point
    //i,j : point
    //in,jn : next point
    int di1,di2,dj1,dj2;
    int k;
    float dx2,dy2,r;
    float h,b,h0,b0;
    dx2=dx*dx; 
    dy2=dy*dy;
    r=dy2/dx2;
    di1 = i-ip;
    di2 = in-i;
    dj1 = j-jp;
    dj2 = jn-j;
    k = (i+1)+largeur*(j+1);
    h = (j+1.0)/(j+0.5);
    b = (j-1.0)/(j-0.5);
    h0 = j/(j+0.5);
    b0 = j/(j-0.5);
    if ((abs(di1+dj1)!=1)||(abs(di2+dj2)!=1)) return POISSON_BAD_INDEX_SPEC;
    if ((di1==di2)&&(dj1==dj2)) { // segment rectiligne
        if (di1==1) {
            D[k] = G[k] =r*(a1+a2)/2; H[k] = a1*h; B[k] = a2*b; C[k] = -(G[k]+D[k]+a1*h0+a2*b0); F[k] = dy2*source;
        }
        else if (di1==-1) {
            D[k] = G[k] = r*(a1+a2)/2; H[k] = a2*h; B[k] = a1*b; C[k] = -(G[k]+D[k]+a2*h0+a1*b0); F[k] = dy2*source;   
        }
        else if (dj1==1) {
            G[k] = a1*r; D[k] = a2*r; H[k] = (a1+a2)/2*h; B[k] = (a1+a2)/2*b; C[k] = -(G[k]+D[k]+(a1+a2)/2*(h0+b0)); F[k] = dy2*source; 
        }
        else if (dj1==-1) {
            G[k] = a2*r; D[k] = a1*r; H[k] = (a1+a2)/2*h; B[k] = (a1+a2)/2*b; C[k] = -(G[k]+D[k]+(a1+a2)/2*(h0+b0)); F[k] = dy2*source;    
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    else { // angle
        if ((di1==1)&&(dj2==1)) {
            G[k] = r*(a1+a2)/2; D[k] = r*a2; B[k] = a2*b; H[k] = (a1+a2)/2*h; C[k] = -(G[k]+D[k]+(a1+a2)/2*h0+a2*b0); F[k] = dy2*source; 
        }
        else if ((di1==-1)&&(dj2==-1)) {
            G[k] = r*a2; D[k] = r*(a1+a2)/2; B[k] = (a1+a2)/2*b; H[k] = a2*h; C[k] = -(G[k]+D[k]+a2*h0+(a1+a2)/2*b0); F[k] = dy2*source;  
        }
        else if ((di2==1)&&(dj1==-1)) {
            G[k] = r*a2; D[k] = r*(a1+a2)/2; B[k] = a2*b; H[k] = (a1+a2)/2*h; C[k] = -(G[k]+D[k]+(a1+a2)/2*h0+a2*b0); F[k] = dy2*source;   
        }
        else if ((di2==-1)&&(dj1==1)) {
             G[k] = r*(a1+a2)/2; D[k] = r*a2; B[k] = (a1+a2)/2*b; H[k] = a2*h; C[k] = -(G[k]+D[k]+a2*h0+(a1+a2)/2*b0); F[k] = dy2*source; 
        }
        else if ((di1==1)&&(dj2==-1)) {
           G[k] = r*(a1+a2)/2; D[k] = r*a1; B[k] = (a1+a2)/2*b; H[k] = a1*h; C[k] = -(G[k]+D[k]+a1*h0+(a1+a2)/2*b0); F[k] = dy2*source; 
        }
        else if ((di1==-1)&&(dj2==1)) {
           G[k] = r*a1; D[k] = r*(a1+a2)/2; B[k] = a1*b; H[k] = (a1+a2)/2*h; C[k] = -(G[k]+D[k]+(a1+a2)/2*h0+a1*b0); F[k] = dy2*source; 
        }
        else if ((dj1==-1)&&(di2==-1)) {
            G[k] = r*(a1+a2)/2; D[k] = r*a1; B[k] = a1*b; H[k] = (a1+a2)/2*h; C[k] = -(G[k]+D[k]+(a1+a2)/2*h0+a1*b0); F[k] = dy2*source; 
        }
        else if ((dj1==1)&&(di2==1)) {
             G[k] = r*a1; D[k] = r*(a1+a2)/2; B[k] = (a1+a2)/2*b; H[k] = a1*h; C[k] = -(G[k]+D[k]+a1*h0+(a1+a2)/2*b0); F[k] = dy2*source;  
        }
        else return POISSON_BAD_INDEX_SPEC;
    }
    
    return 0;       
}

int poisson_neumann_polygon(intArray ilist, intArray jlist, float source, float derivX, float derivY, float dx, float dy) {
    int k,n;
    int err;
    if ((ilist.ndims!=1)||(jlist.ndims!=1)) return POISSON_BAD_ARRAY_DIM;
    if (ilist.dims[0]!=jlist.dims[0]) return POISSON_BAD_ARRAY_DIM;
    n = ilist.dims[0];
    if (n<2) return POISSON_BAD_ARRAY_DIM;
    err=poisson_neumann(ilist.data[n-1],jlist.data[n-1],ilist.data[0],jlist.data[0],ilist.data[1],jlist.data[1],source,derivX,derivY,dx,dy);
    if (err!=0) return err;
    for (k=1; k<n-1; k++) {
        err=poisson_neumann(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source,derivX,derivY,dx,dy);
        if (err!=0) return err;
    }
    err=poisson_neumann(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[0],jlist.data[0],source,derivX,derivY,dx,dy);
    if (err!=0) return err;
    return 0;
}

int poisson_neumann_polygon_axial(intArray ilist, intArray jlist, float source, float derivX, float derivY, float dx, float dy) {
    int k,n;
    int err;
    if ((ilist.ndims!=1)||(jlist.ndims!=1)) return POISSON_BAD_ARRAY_DIM;
    if (ilist.dims[0]!=jlist.dims[0]) return POISSON_BAD_ARRAY_DIM;
    n = ilist.dims[0];
    if (n<2) return POISSON_BAD_ARRAY_DIM;
    err=poisson_neumann_axial(ilist.data[n-1],jlist.data[n-1],ilist.data[0],jlist.data[0],ilist.data[1],jlist.data[1],source,derivX,derivY,dx,dy);
    if (err!=0) return err;
    for (k=1; k<n-1; k++) {
        err=poisson_neumann_axial(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source,derivX,derivY,dx,dy);
        if (err!=0) return err;
    }
    err=poisson_neumann_axial(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[0],jlist.data[0],source,derivX,derivY,dx,dy);
    if (err!=0) return err;
    return 0;
}

int poisson_neumann_polygon_axial_vect(intArray ilist, intArray jlist, float source, float derivX, float derivY, float dx, float dy) {
    int k,n;
    int err;
    if ((ilist.ndims!=1)||(jlist.ndims!=1)) return POISSON_BAD_ARRAY_DIM;
    if (ilist.dims[0]!=jlist.dims[0]) return POISSON_BAD_ARRAY_DIM;
    n = ilist.dims[0];
    if (n<2) return POISSON_BAD_ARRAY_DIM;
    err=poisson_neumann_axial_vect(ilist.data[n-1],jlist.data[n-1],ilist.data[0],jlist.data[0],ilist.data[1],jlist.data[1],source,derivX,derivY,dx,dy);
    if (err!=0) return err;
    for (k=1; k<n-1; k++) {
        err=poisson_neumann_axial_vect(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source,derivX,derivY,dx,dy);
        if (err!=0) return err;
    }
    err=poisson_neumann_axial_vect(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[0],jlist.data[0],source,derivX,derivY,dx,dy);
    if (err!=0) return err;
    return 0;
}


int poisson_limit_polygon(intArray ilist, intArray jlist, intArray limitType, floatArray source, floatArray value, floatArray derivX, floatArray derivY, float dx, float dy) {
    int k,n;
    int err;
   
    if ((ilist.ndims!=1)||(jlist.ndims!=1)||(limitType.ndims!=1)||(source.ndims!=1)||(value.ndims!=1)||(derivX.ndims!=1)||(derivY.ndims!=1)) return POISSON_BAD_ARRAY_DIM;
    if (ilist.dims[0]!=jlist.dims[0]) return POISSON_BAD_ARRAY_DIM;
    n = ilist.dims[0];
    if ((limitType.dims[0]!=n)||(source.dims[0]!=n)||(value.dims[0]!=n)||(derivX.dims[0]!=n)||(derivY.dims[0]!=n)) return POISSON_BAD_ARRAY_DIM;
    k = 0;
    if (limitType.data[k]==LIMIT_SOURCE) {err=poisson_source(ilist.data[k],jlist.data[k],source.data[k],dx,dy); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_DIRICHLET) {err=poisson_dirichlet(ilist.data[k],jlist.data[k],value.data[k]); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_NEUMANN) {err=poisson_neumann(ilist.data[n-1],jlist.data[n-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_INTERFACE) {err=poisson_interface(ilist.data[n-1],jlist.data[n-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    for (k=1; k<n-1; k++) {
          if (limitType.data[k]==LIMIT_SOURCE) {err=poisson_source(ilist.data[k],jlist.data[k],source.data[k],dx,dy); if (err!=0) return err;}
          else if (limitType.data[k]==LIMIT_DIRICHLET) {err=poisson_dirichlet(ilist.data[k],jlist.data[k],value.data[k]); if (err!=0) return err;}
          else if (limitType.data[k]==LIMIT_NEUMANN) {err=poisson_neumann(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
          else if (limitType.data[k]==LIMIT_INTERFACE) {err=poisson_interface(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    }
    if (limitType.data[k]==LIMIT_SOURCE) {err=poisson_source(ilist.data[k],jlist.data[k],source.data[k],dx,dy); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_DIRICHLET) {err=poisson_dirichlet(ilist.data[k],jlist.data[k],value.data[k]); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_NEUMANN) {err=poisson_neumann(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[0],jlist.data[0],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_INTERFACE) {err=poisson_interface(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[0],jlist.data[0],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    return 0;
}

int poisson_limit_polygon_axial(intArray ilist, intArray jlist, intArray limitType, floatArray source, floatArray value, floatArray derivX, floatArray derivY, float dx, float dy) {
    int k,n;
    int err;
    if ((ilist.ndims!=1)||(jlist.ndims!=1)||(limitType.ndims!=1)||(source.ndims!=1)||(value.ndims!=1)||(derivX.ndims!=1)||(derivY.ndims!=1)) return POISSON_BAD_ARRAY_DIM;
    if (ilist.dims[0]!=jlist.dims[0]) return POISSON_BAD_ARRAY_DIM;
    n = ilist.dims[0];
    if ((limitType.dims[0]!=n)||(source.dims[0]!=n)||(value.dims[0]!=n)||(derivX.dims[0]!=n)||(derivY.dims[0]!=n)) return POISSON_BAD_ARRAY_DIM;
    k = 0;
    if (limitType.data[k]==LIMIT_SOURCE) {err=poisson_source_axial(ilist.data[k],jlist.data[k],source.data[k],dx,dy); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_DIRICHLET) {err=poisson_dirichlet(ilist.data[k],jlist.data[k],value.data[k]); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_NEUMANN) {err=poisson_neumann_axial(ilist.data[n-1],jlist.data[n-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_INTERFACE) {err=poisson_interface_axial(ilist.data[n-1],jlist.data[n-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    for (k=1; k<n-1; k++) {
          if (limitType.data[k]==LIMIT_SOURCE) {err=poisson_source_axial(ilist.data[k],jlist.data[k],source.data[k],dx,dy); if (err!=0) return err;}
          else if (limitType.data[k]==LIMIT_DIRICHLET) {err=poisson_dirichlet(ilist.data[k],jlist.data[k],value.data[k]); if (err!=0) return err;}
          else if (limitType.data[k]==LIMIT_NEUMANN) {err=poisson_neumann_axial(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
          else if (limitType.data[k]==LIMIT_INTERFACE) {err=poisson_interface_axial(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    }
    if (limitType.data[k]==LIMIT_SOURCE) {err=poisson_source_axial(ilist.data[k],jlist.data[k],source.data[k],dx,dy); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_DIRICHLET) {err=poisson_dirichlet(ilist.data[k],jlist.data[k],value.data[k]); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_NEUMANN) {err=poisson_neumann_axial(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[0],jlist.data[0],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_INTERFACE) {err=poisson_interface_axial(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[0],jlist.data[0],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    return 0;
}

int poisson_limit_polygon_axial_vect(intArray ilist, intArray jlist, intArray limitType, floatArray source, floatArray value, floatArray derivX, floatArray derivY, float dx, float dy) {
    int k,n;
    int err;
   
    if ((ilist.ndims!=1)||(jlist.ndims!=1)||(limitType.ndims!=1)||(source.ndims!=1)||(value.ndims!=1)||(derivX.ndims!=1)||(derivY.ndims!=1)) return POISSON_BAD_ARRAY_DIM;
    if (ilist.dims[0]!=jlist.dims[0]) return POISSON_BAD_ARRAY_DIM;
    n = ilist.dims[0];
    if ((limitType.dims[0]!=n)||(source.dims[0]!=n)||(value.dims[0]!=n)||(derivX.dims[0]!=n)||(derivY.dims[0]!=n)) return POISSON_BAD_ARRAY_DIM;
    k = 0;
    if (limitType.data[k]==LIMIT_SOURCE) {err=poisson_source(ilist.data[k],jlist.data[k],source.data[k],dx,dy); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_DIRICHLET) {err=poisson_dirichlet(ilist.data[k],jlist.data[k],value.data[k]); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_NEUMANN) {err=poisson_neumann_axial_vect(ilist.data[n-1],jlist.data[n-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_INTERFACE) {err=poisson_interface_axial_vect(ilist.data[n-1],jlist.data[n-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    for (k=1; k<n-1; k++) {
          if (limitType.data[k]==LIMIT_SOURCE) {err=poisson_source(ilist.data[k],jlist.data[k],source.data[k],dx,dy); if (err!=0) return err;}
          else if (limitType.data[k]==LIMIT_DIRICHLET) {err=poisson_dirichlet(ilist.data[k],jlist.data[k],value.data[k]); if (err!=0) return err;}
          else if (limitType.data[k]==LIMIT_NEUMANN) {err=poisson_neumann_axial_vect(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
          else if (limitType.data[k]==LIMIT_INTERFACE) {err=poisson_interface_axial_vect(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[k+1],jlist.data[k+1],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    }
    if (limitType.data[k]==LIMIT_SOURCE) {err=poisson_source(ilist.data[k],jlist.data[k],source.data[k],dx,dy); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_DIRICHLET) {err=poisson_dirichlet(ilist.data[k],jlist.data[k],value.data[k]); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_NEUMANN) {err=poisson_neumann_axial_vect(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[0],jlist.data[0],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    else if (limitType.data[k]==LIMIT_INTERFACE) {err=poisson_interface_axial_vect(ilist.data[k-1],jlist.data[k-1],ilist.data[k],jlist.data[k],ilist.data[0],jlist.data[0],source.data[k],derivX.data[k],derivY.data[k],dx,dy); if (err!=0) return err;}
    return 0;
}

int poisson_neumann_borders(float source, float derivX1, float derivX2, float derivY1, float derivY2, float dx, float dy) {
    int i,j;
    int err;
    err=poisson_neumann(0,1,0,0,1,0,source,derivX1,derivY1,dx,dy); // coin inférieur gauche
    if (err!=0) return err;
    for (i=1; i<width-1; i++) { // bord inférieur
        err=poisson_neumann(i-1,0,i,0,i+1,0,source,0,derivY1,dx,dy);
        if (err!=0) return err;
    }
    err=poisson_neumann(i-1,0,i,0,i,1,source,derivX2,derivY1,dx,dy); // coin inférieur droit
    if (err!=0) return err;
    for (j=1; j<height-1; j++) { // bord droit
        err=poisson_neumann(i,j-1,i,j,i,j+1,source,derivX2,0,dx,dy);
        if (err!=0) return err;
    }
    err=poisson_neumann(i,j-1,i,j,i-1,j,source,derivX2,derivY2,dx,dy); // coin supérieur droit
    if (err!=0) return err;
    for (i=width-2; i>0; i--) { // bord supérieur
        err=poisson_neumann(i+1,j,i,j,i-1,j,source,0,derivY2,dx,dy);
        if (err!=0) return err;
    }
    err=poisson_neumann(i+1,j,i,j,i,j-1,source,derivX1,derivY2,dx,dy); // coin supérieur gauche
    if (err!=0) return err;
    for (j=height-2; j>0; j--) { // bord gauche
        err=poisson_neumann(i,j+1,i,j,i,j-1,source,derivX1,0,dx,dy);
        if (err!=0) return err;
    }
    return 0;
}

int poisson_neumann_borders_axial(float source, float derivX1, float derivX2, float derivY1, float derivY2, float dx, float dy) {
    int i,j,k;
    int err;
    float dx2,dy2,r;
    dx2=dx*dx;
    dy2=dy*dy;
    r=dy2/dx2;
    // coin inférieur gauche
    i = 0; j = 0;
    k = (i+1)+largeur*(j+1);
    G[k] = 0.0; D[k] = r/8.0; H[k] = 0.25; B[k] = 0.0; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = dy2/dx/8.0*derivX1+dy2/16.0*source;
    // coin inférieur droit
    i = width-1;
    k = (i+1)+largeur*(j+1);
    G[k] = r/8.0; D[k] = 0.0; H[k] = 0.25; B[k] = 0.0; C[k] = -(G[k]+D[k]+H[k]+B[k]); F[k] = -dy2/dx/8.0*derivX2+dy2/16.0*source;
    for (j=1; j<height-1; j++) { // bord droit
        err=poisson_neumann_axial(i,j-1,i,j,i,j+1,source,derivX2,0,dx,dy);
        if (err!=0) return err;
    }
    err=poisson_neumann_axial(i,j-1,i,j,i-1,j,source,derivX2,derivY2,dx,dy); // coin supérieur droit
    if (err!=0) return err;
    for (i=width-2; i>0; i--) { // bord supérieur
        err=poisson_neumann_axial(i+1,j,i,j,i-1,j,source,0,derivY2,dx,dy);
        if (err!=0) return err;
    }
    err=poisson_neumann_axial(i+1,j,i,j,i,j-1,source,derivX1,derivY2,dx,dy); // coin supérieur gauche
    if (err!=0) return err;
    for (j=height-2; j>0; j--) { // bord gauche
        err=poisson_neumann_axial(i,j+1,i,j,i,j-1,source,derivX1,0,dx,dy);
        if (err!=0) return err;
    }
    return 0;
}


int poisson_neumann_borders_axial_vect(float source, float derivX1, float derivX2, float derivY1, float derivY2, float dx, float dy) {
    int i,j;
    int err;
    i = width-1;
    for (j=1; j<height-1; j++) { // bord droit
        err=poisson_neumann_axial_vect(i,j-1,i,j,i,j+1,source,derivX2,0,dx,dy);
        if (err!=0) return err;
    }
    err=poisson_neumann_axial_vect(i,j-1,i,j,i-1,j,source,derivX2,derivY2,dx,dy); // coin supérieur droit
    if (err!=0) return err;
    for (i=width-2; i>0; i--) { // bord supérieur
        err=poisson_neumann_axial_vect(i+1,j,i,j,i-1,j,source,0,derivY2,dx,dy);
        if (err!=0) return err;
    }
    err=poisson_neumann_axial_vect(i+1,j,i,j,i,j-1,source,derivX1,derivY2,dx,dy); // coin supérieur gauche
    if (err!=0) return err;
    for (j=height-2; j>0; j--) { // bord gauche
        err=poisson_neumann_axial_vect(i,j+1,i,j,i,j-1,source,derivX1,0,dx,dy);
        if (err!=0) return err;
    }
    return 0;
}

int poisson_sinus_mode(int mx, int my) {
    int i,j;
    float fx = (float)(3.1415926*mx/width);
    float fy = (float)(3.1415926*my/height);
    for (i=1; i<=width; i++) {
        for (j=1; j<=height; j++) {
            U[i+largeur*j] = (float)(sin(fx*i)*sin(fy*j));
            
        }
    }
    return 0;
}

#ifdef OPENCL
int poisson_opencl_platforms() {
    cl_int err;
    int p,k;
    char *name;
    size_t s;
    cl_platform_id* platforms;
    cl_uint numDev;
    cl_device_id* devices;
    err = clGetPlatformIDs(0,NULL,&numPlatforms);
    OPENCL_ERROR(err,"clGetPlatformIDs")
    printf("Nombre de platformes openCL : %d\n",numPlatforms);
    platforms = (cl_platform_id*)malloc(sizeof(cl_platform_id)*numPlatforms);
    err = clGetPlatformIDs(numPlatforms,platforms,NULL);
    OPENCL_ERROR(err,"clGetPlatformIDs")
    for (p=0; p<numPlatforms; p++) {
        err = clGetPlatformInfo(platforms[p],CL_PLATFORM_NAME,0,NULL,&s);
        OPENCL_ERROR(err,"clGetPlatformInfo")
        name = (char*)malloc(sizeof(char)*s);
        err = clGetPlatformInfo(platforms[p],CL_PLATFORM_NAME,s,name,NULL);
        OPENCL_ERROR(err,"clGetPlatformInfo")
        printf("Plateforme OpenCL %d : %s\n",p,name);
        free(name);
        //recherche des périphériques
        err = clGetDeviceIDs(platforms[p],CL_DEVICE_TYPE_ALL,0,NULL,&numDev);
        OPENCL_ERROR(err,"clGetDeviceIDs")
        printf("\t\tNombre de peripheriques : %d\n",numDev);
        devices = (cl_device_id*)malloc(sizeof(cl_device_id)*numDev);
        err = clGetDeviceIDs(platforms[p],CL_DEVICE_TYPE_ALL,numDev,devices,NULL);
        OPENCL_ERROR(err,"clGetDeviceIDs")
        for (k=0; k<numDev; k++) {
            err = clGetDeviceInfo(devices[k],CL_DEVICE_NAME,0,NULL,&s);
            OPENCL_ERROR(err,"clGetDeviceInfo")
            name = (char*)malloc(sizeof(char)*s);
            err = clGetDeviceInfo(devices[k],CL_DEVICE_NAME,s,name,NULL);
            OPENCL_ERROR(err,"clGetDeviceInfo")
            printf("\t\tPeripherique de calcul %d : %s\n",k,name);
            free(name); 
        }
        free(devices);
        printf("\n");
    }
    free(platforms);
    return 0;
}

int poisson_set_opencl_platform_device(int p, int d) {
    cl_int err;
    char *name;
    cl_platform_id* platforms;
    size_t s;
    cl_device_id* devices;
    cl_uint numDev;
    err = clGetPlatformIDs(0,NULL,&numPlatforms);
    OPENCL_ERROR(err,"clGetPlatformIDs")
    platforms = (cl_platform_id*)malloc(sizeof(cl_platform_id)*numPlatforms);
    err = clGetPlatformIDs(numPlatforms,platforms,NULL);
    OPENCL_ERROR(err,"clGetPlatformIDs")
    if ((p<0)||(p>=numPlatforms)) return POISSON_BAD_PLATFORM_NUMBER;
    platform = platforms[p];
    err = clGetPlatformInfo(platform,CL_PLATFORM_NAME,0,NULL,&s);
    OPENCL_ERROR(err,"clGetPlatformInfo")
    name = (char*)malloc(sizeof(char)*s);
    err = clGetPlatformInfo(platform,CL_PLATFORM_NAME,s,name,NULL);
    OPENCL_ERROR(err,"clGetPlatformInfo")
    printf("Plateforme OpenCL choisie : %s\n",name);
    free(name);
    free(platforms);
    err = clGetDeviceIDs(platform,CL_DEVICE_TYPE_ALL,0,NULL,&numDev);
    OPENCL_ERROR(err,"clGetDeviceIDs")
    devices = (cl_device_id*)malloc(sizeof(cl_device_id)*numDev);
    err = clGetDeviceIDs(platform,CL_DEVICE_TYPE_ALL,numDev,devices,NULL);
    if ((d<0)||(d>=numDev)) return POISSON_BAD_DEVICE_NUMBER;
    device = devices[d];
    free(devices);
    err = clGetDeviceInfo(device,CL_DEVICE_NAME,0,NULL,&s);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    name = (char*)malloc(sizeof(char)*s);
    err = clGetDeviceInfo(device,CL_DEVICE_NAME,s,name,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    printf("Peripherique de calcul choisi : %s",name);
    free(name);
    err = clGetDeviceInfo(device,CL_DEVICE_TYPE,sizeof(cl_device_type),&deviceType,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    if (deviceType==CL_DEVICE_TYPE_GPU) printf("\t (GPU)");
    else if (deviceType==CL_DEVICE_TYPE_CPU) printf("\t (CPU)");
    printf("\n");
    return 0;
}

int poisson_opencl_init() {
    int err;
    cl_context_properties context_props[3];
    
    if (platform==0) {
        err = poisson_set_opencl_platform_device(0,0);
        if (err<0) return err;
    }
    
    err = clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_ITEM_DIMENSIONS,sizeof(maxWorkItemDimensions),&maxWorkItemDimensions,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    //printf("Max work group item dimensions : %d\n",maxWorkItemDimensions);
    err = clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_ITEM_SIZES,sizeof(maxWorkItemSizes),maxWorkItemSizes,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    
    //for (d=0; d<maxWorkItemDimensions; d++) printf("Max work item in workgroup for dimension %d : %d\n",d+1,maxWorkItemSizes[d]);
    err = clGetDeviceInfo(device,CL_DEVICE_MAX_WORK_GROUP_SIZE,sizeof(maxWorkGroupSize),&maxWorkGroupSize,NULL);
    OPENCL_ERROR(err,"clGetDeviceInfo")
    //printf("Max work group size : %d\n",maxWorkGroupSize);
    context_props[0] = CL_CONTEXT_PLATFORM;
    context_props[1] = (cl_context_properties)platform;
    context_props[2] = 0;
    context = clCreateContextFromType(context_props,deviceType, NULL, NULL, &err);
    OpenCLProgram = clCreateProgramWithSource(context,SOURCE_NB_LINES,programSource,NULL,NULL);
    OPENCL_ERROR(err,"clCreateProgramWithSource")
    err = clBuildProgram(OpenCLProgram, 0, NULL, NULL, NULL, NULL);
    OPENCL_ERROR(err,"clBuildProgram")
    commandQueue = clCreateCommandQueue(context, device, 0, &err);
    OPENCL_ERROR(err,"clCreateCommandQueue") 
    iterationPaire = clCreateKernel(OpenCLProgram, "iterationPaire",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    iterationImpaire = clCreateKernel(OpenCLProgram, "iterationImpaire",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    arrayNorm = clCreateKernel(OpenCLProgram,"arrayNorm",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    iterationPaireColonne = clCreateKernel(OpenCLProgram,"iterationPaireColonne",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    iterationImpaireColonne = clCreateKernel(OpenCLProgram,"iterationImpaireColonne",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    iterationPaireLigne = clCreateKernel(OpenCLProgram,"iterationPaireLigne",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    iterationImpaireLigne = clCreateKernel(OpenCLProgram,"iterationImpaireLigne",&err);
    OPENCL_ERROR(err,"clCreateKernel")
    return 0;
}
 
int poisson_opencl_create_memory() {
    int err;
    // tableaux GPU
    gpu_G = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*taille,G,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    gpu_D = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*taille,D,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    gpu_H = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*taille,H,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    gpu_B = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*taille,B,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    gpu_C = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*taille,C,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    gpu_F = clCreateBuffer(context,CL_MEM_READ_ONLY |CL_MEM_COPY_HOST_PTR,sizeof(float)*taille,F,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    gpu_U = clCreateBuffer(context,CL_MEM_READ_WRITE |CL_MEM_COPY_HOST_PTR,sizeof(float)*taille,U,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    gpu_V = clCreateBuffer(context,CL_MEM_READ_WRITE |CL_MEM_COPY_HOST_PTR,sizeof(float)*taille,U,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    return 0;
}

int poisson_opencl_iterations(int niter, float w) {
    int err;
    size_t global_work_size[2];
    size_t local_work_size[2];
    size_t global_work_size_colonne[2];
    size_t local_work_size_colonne[2];
    size_t global_work_size_ligne[2];
    size_t local_work_size_ligne[2];
    int x,y,s;
    // offset pour ne pas commencer par les bords gauche et haut
    size_t global_work_offset[2] = {1,1};
    size_t global_work_offset_rangee[1] = {1};
    int n;
    float w1;
    w1 = 1.0-w;
    size_t wg_size, wg_size_mul;
    // arguments des noyaux
    err = clSetKernelArg(iterationPaire, 0, sizeof(cl_mem),(void*)&gpu_D);
    err = clSetKernelArg(iterationPaire, 1, sizeof(cl_mem),(void*)&gpu_G);
    err = clSetKernelArg(iterationPaire, 2, sizeof(cl_mem),(void*)&gpu_H);
    err = clSetKernelArg(iterationPaire, 3, sizeof(cl_mem),(void*)&gpu_B);
    err = clSetKernelArg(iterationPaire, 4, sizeof(cl_mem),(void*)&gpu_C);
    err = clSetKernelArg(iterationPaire, 5, sizeof(cl_mem),(void*)&gpu_F);
    err = clSetKernelArg(iterationPaire, 6, sizeof(cl_mem),(void*)&gpu_U);
    err = clSetKernelArg(iterationPaire, 7, sizeof(largeur),&largeur);
    err = clSetKernelArg(iterationPaire, 8, sizeof(w1),&w1);
    err = clSetKernelArg(iterationPaire, 9, sizeof(w),&w);
    OPENCL_ERROR(err,"clSetKernelArg")
    err = clSetKernelArg(iterationImpaire, 0, sizeof(cl_mem),(void*)&gpu_D);
    err = clSetKernelArg(iterationImpaire, 1, sizeof(cl_mem),(void*)&gpu_G);
    err = clSetKernelArg(iterationImpaire, 2, sizeof(cl_mem),(void*)&gpu_H);
    err = clSetKernelArg(iterationImpaire, 3, sizeof(cl_mem),(void*)&gpu_B);
    err = clSetKernelArg(iterationImpaire, 4, sizeof(cl_mem),(void*)&gpu_C);
    err = clSetKernelArg(iterationImpaire, 5, sizeof(cl_mem),(void*)&gpu_F);
    err = clSetKernelArg(iterationImpaire, 6, sizeof(cl_mem),(void*)&gpu_U);
    err = clSetKernelArg(iterationImpaire, 7, sizeof(largeur),&largeur);
    err = clSetKernelArg(iterationImpaire, 8, sizeof(w1),&w1);
    err = clSetKernelArg(iterationImpaire, 9, sizeof(w),&w);
    OPENCL_ERROR(err,"clSetKernelArg")
    err = clSetKernelArg(iterationPaireColonne, 0, sizeof(cl_mem),(void*)&gpu_D);
    err = clSetKernelArg(iterationPaireColonne, 1, sizeof(cl_mem),(void*)&gpu_G);
    err = clSetKernelArg(iterationPaireColonne, 2, sizeof(cl_mem),(void*)&gpu_H);
    err = clSetKernelArg(iterationPaireColonne, 3, sizeof(cl_mem),(void*)&gpu_B);
    err = clSetKernelArg(iterationPaireColonne, 4, sizeof(cl_mem),(void*)&gpu_C);
    err = clSetKernelArg(iterationPaireColonne, 5, sizeof(cl_mem),(void*)&gpu_F);
    err = clSetKernelArg(iterationPaireColonne, 6, sizeof(cl_mem),(void*)&gpu_U);
    err = clSetKernelArg(iterationPaireColonne, 7, sizeof(largeur),&largeur);
    err = clSetKernelArg(iterationPaireColonne, 8, sizeof(height),&height);
    err = clSetKernelArg(iterationPaireColonne, 9, sizeof(width),&width);
    err = clSetKernelArg(iterationPaireColonne, 10, sizeof(w1),&w1);
    err = clSetKernelArg(iterationPaireColonne, 11, sizeof(w),&w);
    OPENCL_ERROR(err,"clSetKernelArg")
    err = clSetKernelArg(iterationImpaireColonne, 0, sizeof(cl_mem),(void*)&gpu_D);
    err = clSetKernelArg(iterationImpaireColonne, 1, sizeof(cl_mem),(void*)&gpu_G);
    err = clSetKernelArg(iterationImpaireColonne, 2, sizeof(cl_mem),(void*)&gpu_H);
    err = clSetKernelArg(iterationImpaireColonne, 3, sizeof(cl_mem),(void*)&gpu_B);
    err = clSetKernelArg(iterationImpaireColonne, 4, sizeof(cl_mem),(void*)&gpu_C);
    err = clSetKernelArg(iterationImpaireColonne, 5, sizeof(cl_mem),(void*)&gpu_F);
    err = clSetKernelArg(iterationImpaireColonne, 6, sizeof(cl_mem),(void*)&gpu_U);
    err = clSetKernelArg(iterationImpaireColonne, 7, sizeof(largeur),&largeur);
    err = clSetKernelArg(iterationImpaireColonne, 8, sizeof(height),&height);
    err = clSetKernelArg(iterationImpaireColonne, 9, sizeof(width),&width);
    err = clSetKernelArg(iterationImpaireColonne, 10, sizeof(w1),&w1);
    err = clSetKernelArg(iterationImpaireColonne, 11, sizeof(w),&w);
    OPENCL_ERROR(err,"clSetKernelArg")
    err = clSetKernelArg(iterationPaireLigne, 0, sizeof(cl_mem),(void*)&gpu_D);
    err = clSetKernelArg(iterationPaireLigne, 1, sizeof(cl_mem),(void*)&gpu_G);
    err = clSetKernelArg(iterationPaireLigne, 2, sizeof(cl_mem),(void*)&gpu_H);
    err = clSetKernelArg(iterationPaireLigne, 3, sizeof(cl_mem),(void*)&gpu_B);
    err = clSetKernelArg(iterationPaireLigne, 4, sizeof(cl_mem),(void*)&gpu_C);
    err = clSetKernelArg(iterationPaireLigne, 5, sizeof(cl_mem),(void*)&gpu_F);
    err = clSetKernelArg(iterationPaireLigne, 6, sizeof(cl_mem),(void*)&gpu_U);
    err = clSetKernelArg(iterationPaireLigne, 7, sizeof(largeur),&largeur);
    err = clSetKernelArg(iterationPaireLigne, 8, sizeof(height),&height);
    err = clSetKernelArg(iterationPaireLigne, 9, sizeof(width),&width);
    err = clSetKernelArg(iterationPaireLigne, 10, sizeof(w1),&w1);
    err = clSetKernelArg(iterationPaireLigne, 11, sizeof(w),&w);
    OPENCL_ERROR(err,"clSetKernelArg")
    err = clSetKernelArg(iterationImpaireLigne, 0, sizeof(cl_mem),(void*)&gpu_D);
    err = clSetKernelArg(iterationImpaireLigne, 1, sizeof(cl_mem),(void*)&gpu_G);
    err = clSetKernelArg(iterationImpaireLigne, 2, sizeof(cl_mem),(void*)&gpu_H);
    err = clSetKernelArg(iterationImpaireLigne, 3, sizeof(cl_mem),(void*)&gpu_B);
    err = clSetKernelArg(iterationImpaireLigne, 4, sizeof(cl_mem),(void*)&gpu_C);
    err = clSetKernelArg(iterationImpaireLigne, 5, sizeof(cl_mem),(void*)&gpu_F);
    err = clSetKernelArg(iterationImpaireLigne, 6, sizeof(cl_mem),(void*)&gpu_U);
    err = clSetKernelArg(iterationImpaireLigne, 7, sizeof(largeur),&largeur);
    err = clSetKernelArg(iterationImpaireLigne, 8, sizeof(height),&height);
    err = clSetKernelArg(iterationImpaireLigne, 9, sizeof(width),&width);
    err = clSetKernelArg(iterationImpaireLigne, 10, sizeof(w1),&w1);
    err = clSetKernelArg(iterationImpaireLigne, 11, sizeof(w),&w);
    OPENCL_ERROR(err,"clSetKernelArg")
    // espace de travail : indices pour accéder aux tableaux (sans les bords)
    
    global_work_size[0] = width-1;
    global_work_size[1] = height-1;
    x = width-1; y = height-1;
    s = x*y;
    err = clGetKernelWorkGroupInfo(iterationPaire,device,CL_KERNEL_WORK_GROUP_SIZE,sizeof(wg_size),&wg_size,NULL);
    OPENCL_ERROR(err,"clGetKernelWorkGroupInfo");
    err = clGetKernelWorkGroupInfo(iterationPaire,device,CL_KERNEL_PREFERRED_WORK_GROUP_SIZE_MULTIPLE,sizeof(wg_size_mul),&wg_size,NULL);
    OPENCL_ERROR(err,"clGetKernelWorkGroupInfo");
    //printf("wg size : %d\n",wg_size);
    //printf("wg size multiple : %d\n",wg_size_mul);
    while (s > maxWorkGroupSize) { x /= 2; y /= 2; s = x*y;}
    local_work_size[0] = x;
    local_work_size[1] = y;
    //printf("group size : %d\n",s);
    
    //dernière colonne et dernière ligne
    global_work_size_colonne[0] = height-1;
    y = height-1;
    while (y > maxWorkGroupSize) {y /= 2;}
    local_work_size_colonne[0] = y;
    global_work_size_ligne[0] = width-1;
    x = width-1;
    while (x > maxWorkGroupSize) {x /= 2;}
    local_work_size_ligne[0] = x;
    // boucle d'itération
    
    for (n=0; n<niter; n++) {
        err = clEnqueueNDRangeKernel(commandQueue,iterationPaire,2,global_work_offset,global_work_size,local_work_size,0,NULL,NULL);
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel")
        err = clEnqueueNDRangeKernel(commandQueue,iterationPaireLigne,1,global_work_offset_rangee,global_work_size_ligne,local_work_size_ligne,0,NULL,NULL); // dernière ligne
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel")
        err = clEnqueueNDRangeKernel(commandQueue,iterationPaireColonne,1,global_work_offset_rangee,global_work_size_colonne,local_work_size_colonne,0,NULL,NULL); // dernière colonne
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel")
        err = clEnqueueBarrier(commandQueue);
        OPENCL_ERROR(err,"clEnqueueBarrier");
        err = clEnqueueNDRangeKernel(commandQueue,iterationImpaire,2,global_work_offset,global_work_size,local_work_size,0,NULL,NULL);
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel")
        err = clEnqueueNDRangeKernel(commandQueue,iterationImpaireLigne,1,global_work_offset_rangee,global_work_size_ligne,local_work_size_ligne,0,NULL,NULL); // dernière ligne
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel")
        err = clEnqueueNDRangeKernel(commandQueue,iterationImpaireColonne,1,global_work_offset_rangee,global_work_size_colonne,local_work_size_colonne,0,NULL,NULL); // dernière colonne
        OPENCL_ERROR(err,"clEnqueueNDRangeKernel")
        err = clEnqueueBarrier(commandQueue);
        OPENCL_ERROR(err,"clEnqueueBarrier")
    }
    // récupération du tableau U
    err = clEnqueueReadBuffer(commandQueue, gpu_U, CL_TRUE, 0,taille*sizeof(float), U, 0, NULL,NULL);
    OPENCL_ERROR(err,"clEnqueueReadBuffer")
    /*int i,j,k;
    for (j=1; j<=height; j++) {
        for (i=1; i<=width; i++) {
            k = i+largeur*j;
            printf("%f ",U[k]);
        }
        printf("\n");
    }*/
    return 0;
}



int poisson_opencl_get_array_norm(float *value) {
    int err;
    int workGroupSize;
    int numberOfGroups;
    short w,h,s;
    size_t global_work_size[1];
    size_t local_work_size[1];
    size_t src_origin[3] = {sizeof(float),1,0};
    size_t dst_origin[3] = {0,0,0};
    size_t region[3] = {(width-1)*sizeof(float),height-1,1};
    size_t src_row_pitch = largeur*sizeof(float);
    size_t src_slice_pitch = 0;
    size_t dst_row_pitch = (width-1)*sizeof(float);
    size_t dst_slice_pitch = 0;
    int k; 
    float n = 0.0;
    float *norm;
    cl_mem gpu_U1;
    cl_mem gpu_norm;
    
    w = width-1; //largeur puissance de 2, on ignore la dernière colonne dans le calcul de la norme
    h = height-1; // on ignore la dernière ligne
    s=w*h;
    
    if (maxWorkGroupSize < s) {
        workGroupSize = maxWorkGroupSize;
        numberOfGroups = s / workGroupSize;
    }
    else {
        workGroupSize = s;
        numberOfGroups = 1;
    }
    
    
    gpu_U1 = clCreateBuffer(context,CL_MEM_READ_WRITE,sizeof(float)*s,NULL,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    gpu_norm = clCreateBuffer(context,CL_MEM_WRITE_ONLY,sizeof(float)*numberOfGroups,NULL,&err);
    OPENCL_ERROR(err,"clCreateBuffer")
    err = clSetKernelArg(arrayNorm, 0, sizeof(cl_mem),(void*)&gpu_U1);
    err = clSetKernelArg(arrayNorm,1,sizeof(float)*workGroupSize,NULL);
    err = clSetKernelArg(arrayNorm,2,sizeof(cl_mem),(void*)&gpu_norm);
    OPENCL_ERROR(err,"clSetKernelArg")
    err = clEnqueueCopyBufferRect(commandQueue,gpu_U,gpu_U1,src_origin,dst_origin,region,src_row_pitch,src_slice_pitch,dst_row_pitch,dst_slice_pitch,0,NULL,NULL);
    OPENCL_ERROR(err,"clEnqueueCopyBufferRect")
    err = clEnqueueBarrier(commandQueue);
    OPENCL_ERROR(err,"clEnqueueBarrier");
    global_work_size[0] = size;
    local_work_size[0] = workGroupSize;    
    err = clEnqueueNDRangeKernel(commandQueue,arrayNorm,1,NULL,global_work_size,local_work_size,0,NULL,NULL);
    OPENCL_ERROR(err,"clEnqueueNDRangeKernel")
    err = clEnqueueBarrier(commandQueue);
    OPENCL_ERROR(err,"clEnqueueBarrier");
    norm = (float*)malloc(sizeof(float)*numberOfGroups);
    err = clEnqueueReadBuffer(commandQueue, gpu_norm, CL_TRUE, 0,numberOfGroups*sizeof(float), norm, 0, NULL,NULL);
    OPENCL_ERROR(err,"clEnqueueReadBuffer");
    
    for (k=0; k<numberOfGroups; k++) n+= norm[k];
    *value = n/s;
    free(norm);
    clReleaseMemObject(gpu_U1);
    clReleaseMemObject(gpu_norm);
    return 0;
}
int poisson_opencl_release_memory() {
    clReleaseMemObject(gpu_G);
    clReleaseMemObject(gpu_D);
    clReleaseMemObject(gpu_H);
    clReleaseMemObject(gpu_B);
    clReleaseMemObject(gpu_C);
    clReleaseMemObject(gpu_F);
    clReleaseMemObject(gpu_U);
    clReleaseMemObject(gpu_V);
    return 0;
}

int poisson_opencl_release() {
    clReleaseKernel(iterationPaire);
    clReleaseKernel(iterationImpaire);
    clReleaseProgram(OpenCLProgram);
    clReleaseCommandQueue(commandQueue);
    clReleaseContext(context);
    return 0;
}
#endif

void poisson_iterationPaire(int jStart, int jEnd, float w1, float w) {
    int i,j,c;
    c = 1+largeur*jStart;
    for (j=jStart; j<=jEnd; j++) {
        for (i=1; i<=width; i++) {
             if ((((i&1)+(j&1))&1)==0) U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];
             c += 1;
        }
        c += 2;
    }      
}

void poisson_iterationImpaire(int jStart, int jEnd, float w1, float w) {
    int i,j,c;
    c = 1+largeur*jStart;
    for (j=jStart; j<=jEnd; j++) {
        for (i=1; i<=width; i++) {
             if ((((i&1)+(j&1))&1)==1) U[c] = w1*U[c]+w*(F[c]-D[c]*U[c+1]-G[c]*U[c-1]-H[c]*U[c+largeur]-B[c]*U[c-largeur])/C[c];
             c += 1;
        }
        c += 2;
    }     
} 

int poisson_iterations(int niter, float w) { // CPU iteration, 1 thread
    int i;
    float w1;
    w1 = 1.0-w;
    for (i=0; i<niter; i++) {
        poisson_iterationPaire(1,height,w1,w);
        poisson_iterationImpaire(1,height,w1,w);
    }
    return 0;
}

void* _iterations_thread_paire(void *arg) {
    threadArg* ta = (threadArg*)arg;
    int niter = ta->niter;
    int jStart = ta->jStart;
    int jEnd = ta->jEnd;
    float w1 = ta->w1;
    float w = ta->w;
    int i,r;
    
    for (i=0; i<niter; i++) {
         poisson_iterationPaire(jStart,jEnd,w1,w);
         r =  pthread_barrier_wait(&threadBarrier);
         if(r != 0 && r != PTHREAD_BARRIER_SERIAL_THREAD) printf("Impossible d'attendre la barriere\n");
    }
    pthread_exit(NULL);
}

void* _iterations_thread_impaire(void *arg) {
    threadArg* ta = (threadArg*)arg;
    int niter = ta->niter;
    int jStart = ta->jStart;
    int jEnd = ta->jEnd;
    float w1 = ta->w1;
    float w = ta->w;
    int i,r;
    
    for (i=0; i<niter; i++) {
         poisson_iterationImpaire(jStart,jEnd,w1,w);
         r =  pthread_barrier_wait(&threadBarrier);
         if(r != 0 && r != PTHREAD_BARRIER_SERIAL_THREAD) printf("Impossible d'attendre la barriere\n");
    }
    pthread_exit(NULL);
}

int poisson_iterations_2threads(int niter, float w) { // CPU iteration, 2 threads
    pthread_attr_t attr;
    pthread_t thread1, thread2;
    threadArg arg1, arg2;
    int err;
    float w1;
    w1 = 1.0-w;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    err = pthread_barrier_init(&threadBarrier, NULL,2);
    if (err) return POISSON_ERROR_THREAD_BARRIER;
    arg1.niter = niter;
    arg1.jStart = 1;
    arg1.jEnd = height;
    arg1.w1 = w1;
    arg1.w = w;
    err = pthread_create(&thread1,&attr,_iterations_thread_paire,(void*)&arg1);
    if (err) return POISSON_ERROR_THREAD_START;
    arg2.niter = niter;
    arg2.jStart = 1;
    arg2.jEnd = height;
    arg2.w1 = w1;
    arg2.w = w;
    err = pthread_create(&thread2,&attr,_iterations_thread_impaire,(void*)&arg2);
    if (err) return POISSON_ERROR_THREAD_START;
    err = pthread_join(thread1,NULL);
    if (err) return POISSON_ERROR_THREAD_JOIN;
    err = pthread_join(thread2,NULL);
    if (err) return POISSON_ERROR_THREAD_JOIN;
    pthread_attr_destroy(&attr);
    pthread_barrier_destroy(&threadBarrier);
    return 0; 
}

int poisson_iterations_4threads(int niter,float w) {
    pthread_attr_t attr;
    pthread_t thread1, thread2, thread3, thread4;
    threadArg arg1, arg2, arg3, arg4;
    int err;
    int h2;
    float w1;
    w1 = 1.0-w;
    if (height<2) poisson_iterations_2threads(niter,w);
    h2 = height/2;
    pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
    err = pthread_barrier_init(&threadBarrier, NULL,4);
    if (err) return POISSON_ERROR_THREAD_BARRIER;
    arg1.niter = niter;
    arg1.jStart = 1;
    arg1.jEnd = h2;
    arg1.w1 = w1;
    arg1.w = w;
    arg2.niter = niter;
    arg2.jStart = h2+1;
    arg2.jEnd = height;
    arg2.w1 = w1;
    arg2.w = w;
    arg3.niter = niter;
    arg3.jStart = 1;
    arg3.jEnd = h2;
    arg3.w1 = w1;
    arg3.w = w;
    arg4.niter = niter;
    arg4.jStart = h2+1;
    arg4.jEnd = height;
    arg4.w1 = w1;
    arg4.w = w;
    err = pthread_create(&thread1,&attr,_iterations_thread_paire,(void*)&arg1);
    if (err) return POISSON_ERROR_THREAD_START;
    err = pthread_create(&thread2,&attr,_iterations_thread_paire,(void*)&arg2);
    if (err) return POISSON_ERROR_THREAD_START;
    err = pthread_create(&thread3,&attr,_iterations_thread_impaire,(void*)&arg3);
    if (err) return POISSON_ERROR_THREAD_START;
    err = pthread_create(&thread4,&attr,_iterations_thread_impaire,(void*)&arg4);
    if (err) return POISSON_ERROR_THREAD_START;
    err = pthread_join(thread1,NULL);
    if (err) return POISSON_ERROR_THREAD_JOIN;
    err = pthread_join(thread2,NULL);
    if (err) return POISSON_ERROR_THREAD_JOIN;
    err = pthread_join(thread3,NULL);
    if (err) return POISSON_ERROR_THREAD_JOIN;
    err = pthread_join(thread4,NULL);
    if (err) return POISSON_ERROR_THREAD_JOIN;
    pthread_attr_destroy(&attr);
    pthread_barrier_destroy(&threadBarrier);
    return 0;
}

int poisson_get_array(floatArray *array, int symetry) {
    int i,j,k;
    int offset;
    if (array->ndims!=2) return -255;
    if (symetry) offset = width*(height-1); 
    else offset = 0;    
    for (j=0; j<height; j++) {
        for (i=0; i<width; i++) {
            k = (i+1)+largeur*(j+1);
            U1[offset+i+j*width] = U[k];
            if ((symetry)&&(j!=0)) U1[offset+i-j*width] = symetry*U[k];
        }
    }
    array->dims[1] = width;
    array->dims[0] = height;
    if (symetry) array->dims[0] = 2*height-1;
    array->data = U1;
    return 0;   
}



int poisson_get_derivX(floatArray *array,float dx, int symetry) {
    int i,j,k;
    int offset;
    if (array->ndims!=2) return -255;
    if (symetry) offset = width*(height-1);  
    else offset = 0;
    i = 0;
    for (j=0; j<height; j++) {
        k = (i+1)+largeur*(j+1);
        U1[offset+i+width*j] = (U[k+1]-U[k])/dx;
        
    }
    for (i=1; i<width-1; i++) {
        for (j=0; j<height; j++) {
            k = (i+1)+largeur*(j+1);
            U1[offset+i+width*j] = (U[k+1]-U[k-1])/(2*dx);
            if (symetry) U1[offset+i-width*j] = symetry*U1[offset+i+width*j];
        }   
    }
    i = width-1;
    for (j=0; j<height; j++) {
        k = (i+1)+largeur*(j+1);
        U1[offset+i+width*j] = (U[k]-U[k-1])/dx;
        if (symetry) U1[offset+i-width*j] = symetry*U1[offset+i+width*j];
    }
    array->dims[1] = width;
    array->dims[0] = height;
    if (symetry) array->dims[0] = 2*height-1;
    array->data = U1;
    return 0;    
}
int poisson_get_derivY(floatArray *array,float dy, int symetry) {
    int i,j,k;
    int offset;
    if (array->ndims!=2) return -255;
    if (symetry) offset = width*(height-1);  
    else offset = 0;
    j=0;
    for (i=0; i<width; i++) {
        k = (i+1)+largeur*(j+1);
        U1[offset+i+width*j] = (U[k+largeur]-U[k])/dy;
        
    }
    for (j=1; j<height-1; j++) {
        for (i=0; i<width; i++) {
            k = (i+1)+largeur*(j+1);
            U1[offset+i+width*j] = (U[k+largeur]-U[k-largeur])/(2*dy);
            if (symetry) U1[offset+i-width*j] = symetry*U1[offset+i+width*j];
        }
    }
    j= height-1;
    for (i=0; i<width; i++) {
        k = (i+1)+largeur*(j+1);
        U1[offset+i+width*j] = (U[k]-U[k-largeur])/dy;
        if (symetry) U1[offset+i-width*j] = symetry*U1[offset+i+width*j];
    }
    array->dims[1] = width;
    array->dims[0] = height;
    if (symetry) array->dims[0] = 2*height-1;
    array->data = U1;
    return 0;   
}

int poisson_get_derivRUR_axial(floatArray *array, float dr, int symetry) {
    int i,j,k;
    int offset;
    if (array->ndims!=2) return -255;
    if (symetry) offset = width*(height-1);  
    else offset = 0;
    j = 0;
    for (i=0; i<width; i++) {
        k=(i+1)+largeur*(j+1);
        U1[offset+i+width*j] = ((j+2)*U[k+largeur]-(j+1)*U[k])/((j+1)*dr);
        
    }
    for (j=1; j<height-1; j++) {
        for (i=0; i<width; i++) {
            k=(i+1)+largeur*(j+1);
            U1[offset+i+width*j] = ((j+1)*U[k+largeur]-(j-1)*U[k-largeur])/(2*j*dr);
            if (symetry) U1[offset+i-width*j] = symetry*U1[offset+i+width*j];
        }   
    }
    j=height-1;
    for (i=0; i<width; i++) {
        k = (i+1)+largeur*(j+1);
        U1[offset+i+width*j] = (j*U[k]-(j-1)*U[k-largeur])/(j*dr);
        if (symetry) U1[offset+i-width*j] = symetry*U1[offset+i+width*j];
    }
    array->dims[1] = width;
    array->dims[0] = height;
    if (symetry) array->dims[0] = 2*height-1;
    array->data = U1;
    return 0; 
}

int poisson_get_array_norm(float *value) {
    int i,j,c;
    float norm = 0;
    float u;
    c = 1+largeur;
    for (j=1; j<=height; j++) {
        for (i=1; i<=width; i++) {
             u = U[c];
             norm += u*u;
             c += 1;
        }
        c += 2;
    }
    *value = norm/size;
    return 0;
}




