#include "dataTypes.h"

#define OPENCL

#define POISSON_KERNEL_FILE_NOT_FOUND -70
#define POISSON_INDEX_OUT_OF_RANGE -71
#define POISSON_BAD_ARRAY_DIM -72
#define POISSON_BAD_INDEX_SPEC -73
#define POISSON_BAD_PLATFORM_NUMBER -74
#define POISSON_UNKNOWN_DEVICE -75
#define POISSON_BAD_DEVICE_NUMBER -76
#define POISSON_ERROR_THREAD_START -77
#define POISSON_ERROR_THREAD_JOIN -78
#define POISSON_ERROR_THREAD_BARRIER -79

#define LIMIT_SOURCE 0
#define LIMIT_DIRICHLET 1
#define LIMIT_NEUMANN 2
#define LIMIT_INTERFACE 3

int poisson_open(int pw, int ph);
int poisson_close(void);
int poisson_laplacien(float dx, float dy);
int poisson_laplacien_axial(float dz, float dr);
int poisson_laplacien_axial_vect(float dz, float dr);
int poisson_dirichlet_borders(float value);
int poisson_dirichlet_borders_axial(float value);
int poisson_dirichlet_borders_axial_vect(float value);
int poisson_sinus_mode(int mx, int my);
int poisson_opencl_platforms(void);
int poisson_set_opencl_platform_device(int p,int d);
int poisson_opencl_init(void);
int poisson_opencl_create_memory(void);
int poisson_opencl_iterations(int niter, float w);
int poisson_opencl_release_memory(void);
int poisson_opencl_release(void);
int poisson_opencl_array(floatArray* array);
int poisson_opencl_get_array_norm(float *value);
void poisson_iterationPaire(int jStart, int jEnd, float w1, float w);
void poisson_iterationImpaire(int jStart, int jEnd, float w1, float w);
int poisson_iterations_2threads(int niter, float w);
int poisson_iterations_4threads(int niter, float w);
int poisson_iterations(int niter, float w);
int poisson_get_array(floatArray *array, int symetry);
int poisson_get_derivX(floatArray *array,float dx, int symetry);
int poisson_get_derivY(floatArray *array,float dy, int symetry);
int poisson_get_derivRUR_axial(floatArray *array,float dr, int symetry);
int poisson_get_array_norm(float *value);
int poisson_dirichlet(int i, int j, float value);
int poisson_dirichlet_polygon(intArray ilist, intArray jlist,float value);
int poisson_source(int i, int j, float s, float dx, float dy);
int poisson_source_axial(int i, int j, float s, float dz, float dr);
int poisson_source_polygon(intArray ilist, intArray jlist, float s, float dx, float dy);
int poisson_source_polygon_axial(intArray ilist, intArray jlist, float s, float dz, float dr);
int poisson_neumann(int ip, int jp, int i, int j, int in, int jn, float source, float derivX, float derivY, float dx, float dy);
int poisson_neumann_axial(int ip, int jp, int i, int j, int in, int jn, float source, float derivX, float derivY, float dx, float dy);
int poisson_neumann_axial_vect(int ip, int jp, int i, int j, int in, int jn, float source, float derivX, float derivY, float dx, float dy);
int poisson_interface(int ip, int jp, int i, int j, int in, int jn, float source,float a1, float a2, float dx, float dy);
int poisson_interface_axial(int ip, int jp, int i, int j, int in, int jn, float source,float a1, float a2, float dx, float dy);
int poisson_interface_axial_vect(int ip, int jp, int i, int j, int in, int jn, float source,float a1, float a2, float dx, float dy);
int poisson_neumann_polygon(intArray ilist, intArray jlist, float source, float derivX, float derivY, float dx, float dy);
int poisson_neumann_polygon_axial(intArray ilist, intArray jlist, float source, float derivX, float derivY, float dx, float dy);
int poisson_neumann_polygon_axial_vect(intArray ilist, intArray jlist, float source, float derivX, float derivY, float dx, float dy);
int poisson_neumann_borders(float source, float derivX1, float derivX2, float derivY1, float derivY2, float dx, float dy);
int poisson_neumann_borders_axial(float source, float derivX1, float derivX2, float derivY1, float derivY2, float dx, float dy);
int poisson_neumann_borders_axial_vect(float source, float derivX1, float derivX2, float derivY1, float derivY2, float dx, float dy);
int poisson_limit_polygon(intArray ilist, intArray jlist, intArray limitType, floatArray source, floatArray value, floatArray derivX, floatArray derivY, float dx, float dy);
int poisson_limit_polygon_axial(intArray ilist, intArray jlist, intArray limitType, floatArray source, floatArray value, floatArray derivX, floatArray derivY, float dx, float dy);
int poisson_limit_polygon_axial_vect(intArray ilist, intArray jlist, intArray limitType, floatArray source, floatArray value, floatArray derivX, floatArray derivY, float dx, float dy);
