#include "dataTypes.h"

//#define OPENCL

#define POISSON_KERNEL_FILE_NOT_FOUND -70
#define POISSON_INDEX_OUT_OF_RANGE -71
#define POISSON_BAD_ARRAY_DIM -72
#define POISSON_BAD_INDEX_SPEC -73
#define POISSON_BAD_PLATFORM_NUMBER -74
#define POISSON_UNKNOWN_DEVICE -75
#define POISSON_BAD_DEVICE_NUMBER -76
#define POISSON_ERROR_THREAD_START -77
#define POISSON_ERROR_THREAD_JOIN -78
#define POISSON_ERROR_THREAD_BARRIER -79
#define POISSON_ERROR_NOT_ENOUGH_GRIDS -80
#define POISSON_TOLERANCE_NOT_REACHED -81
#define POISSON_BAD_POLYGON_DEFINITION -82

#define LIMIT_SOURCE 0
#define LIMIT_DIRICHLET 1
#define LIMIT_NEUMANN 2
#define LIMIT_INTERFACE 3

#define TYPE_CARTESIAN 10
#define TYPE_AXIAL 11
#define TYPE_AXIAL_VECT 12
#define TYPE_POLAR 13

#define RESTRICTION_INJECTION 20
#define RESTRICTION_HALFWEIGHTING 21
#define RESTRICTION_FULLWEIGHTING 22

#define MASK_INSIDE 0
#define MASK_NEUMANN 1
#define MASK_DIRICHLET 2
#define MASK_SOURCE 3
#define MASK_INTERFACE 4
#define MASK_OUTSIDE 5

int poissonMG_open(char type, int pw_min, int ph_min, int levels, float w, float h, char restriction);
int poissonMG_close(void);
int poissonMG_laplacien(void);
int poissonMG_source(int i, int j, float s);
int poissonMG_source_polygon(int i_start, int j_start, intArray li_list, intArray lj_list, float value);
int poissonMG_source_rect(int cx, int cy, int width, int height, float s);
int poissonMG_dirichlet(int i, int j, float value);
int poissonMG_dirichlet_borders(float value);
int poissonMG_dirichlet_polygon(int i_start, int j_start, intArray li_list, intArray lj_list, float value);
int poissonMG_neumann_borders(float source, float derivX1, float derivX2, float derivY1, float derivY2 ) ;
int poissonMG_neumann_polygon(int i_start, int j_start, intArray li_list, intArray lj_list, floatArray derivX_list, floatArray derivY_list, float source);
int poissonMG_neumann_border_ext_polar(float source, float derivR);
int poissonMG_neumann_border_int_polar(int j, float source, float derivR);
int poissonMG_interface_polygon(int i_start, int j_start, intArray li_list, intArray lj_list, float source, float a1, float a2);
int poissonMG_sinus_mode(int mx, int my);
int poissonMG_step_mode(void);
int poissonMG_init_array(void);
int poissonMG_finegrid_iterations(int niter, float w);
int poissonMG_finegrid_iterations_2threads(int niter, float w);
int poissonMG_finegrid_iterations_4threads(int niter, float w);
int poissonMG_get_array(floatArray *array, int symetry);
int poissonMG_get_mask(byteArray *array, int symetry);
int poissonMG_get_derivX(floatArray *array, int symetry);
int poissonMG_get_derivY(floatArray *array, int symetry);
int poissonMG_get_derivRUR_axial(floatArray *array, int symetry);
int poissonMG_get_derivTheta_polar(floatArray *array);
int poissonMG_get_finegrid_residu_norm(float *value);
int poissonMG_get_finegrid_array_norm(float *value);
int poissonMG_multigrid_iterations(int nsteps, int nlevels, intArray niter_list, float w);
int poissonMG_multigrid_full(int mcycle, int nlevels, int niter, int nu1, int nnu2, float *wu);
int poissonMG_multigrid_Vcycles(int nsteps, int nlevels, int niter, int nu1, int nu2);

#ifdef OPENCL
int poissonMG_opencl_platforms(void);
int poissonMG_set_opencl_platform_device(int p, int d);
int poissonMG_opencl_init(void);
int poissonMG_opencl_create_memory_finegrid(void);
int poissonMG_opencl_iterations_finegrid(int niter, float w);
int poissonMG_opencl_get_array_norm_finegrid(float *value);
int poissonMG_opencl_release_memory_finegrid(void);
int poissonMG_opencl_release(void);
#endif

