# -*- coding: utf-8 -*-
"""
Created on Sat Dec 28 19:07:41 2013

@author: fred
"""

from matplotlib.pyplot import *
import math
import numpy
from scipy.interpolate import RectBivariateSpline
from scipy.integrate import odeint

Atheta = numpy.loadtxt("bobine-Atheta.txt")
AthetaDr = numpy.loadtxt("bobine-AthetaDr.txt")
AthetaDz = numpy.loadtxt("bobine-AthetaDz.txt")
Br = numpy.loadtxt("bobine-Br.txt")
Bz = numpy.loadtxt("bobine-Bz.txt")
r = numpy.loadtxt("bobine-r.txt")
z = numpy.loadtxt("bobine-z.txt")
f_A=RectBivariateSpline(z,r,Atheta.transpose())
f_ADr=RectBivariateSpline(z,r,AthetaDr.transpose())
f_ADz=RectBivariateSpline(z,r,AthetaDz.transpose())
f_Bz=RectBivariateSpline(z,r,Bz.transpose())

figure(figsize=(8,8))
streamplot(z,r,Bz,Br,density=3)

alpha=700.0
r0=0.0
z0=-0.7
ptheta = alpha*r0*f_A(z0,r0)
def equation(y,t):
    r=y[0]
    z=y[1]
    A = f_A(z,r)
    if ptheta==0:
        u = -alpha*A
        dpr = u*alpha*f_ADr(z,r)
        dpz = u*alpha*f_ADz(z,r)
        dtheta = -alpha*(f_Bz(z,r)-f_ADr(z,r))
    else:
        u = ptheta/r-alpha*A
        dpr = u*(ptheta/(r*r)+alpha*f_ADr(z,r))
        dpz = u*alpha*f_ADz(z,r)
        dtheta = ptheta/(r*r)-alpha*(f_Bz(z,r)-f_ADr(z,r))
    return [y[2],y[3],dpr,dpz,dtheta]

def trajectoire(r0,z0,pr0,pz0,tmax):
    global ptheta
    ptheta = alpha*r0*f_A(z0,r0)
    te=0.001
    t=numpy.arange(0.0,tmax,te)
    y = odeint(equation,[r0,z0,pr0,pz0,0.0],t,rtol=1e-5,atol=1e-5)
    yt = y.transpose()
    rt = yt[0]
    zt = yt[1]
    plot(zt,rt)
    
def projection(r0,z0,pr0,pz0,tmax):
    global ptheta
    ptheta = alpha*r0*f_A(z0,r0)
    te=0.0001
    t=numpy.arange(0.0,tmax,te)
    y = odeint(equation,[r0,z0,pr0,pz0,0.0],t,rtol=1e-5,atol=1e-5)
    yt = y.transpose()
    rt = yt[0]
    zt = yt[1]
    thetat = yt[4]
    plot(rt*numpy.cos(thetat),rt*numpy.sin(thetat))

def angle(r0,z0,pr0,pz0,tmax):
    global ptheta
    ptheta = alpha*r0*f_A(z0,r0)
    te=0.0001
    t=numpy.arange(0.0,tmax,te)
    y = odeint(equation,[r0,z0,pr0,pz0,0.0],t,rtol=1e-5,atol=1e-5)
    yt = y.transpose()
    rt = yt[0]
    zt = yt[1]
    thetat = yt[4]
    plot(zt,thetat)
    
pz0=0.85
z0 = -0.7
r0=0.00
trajectoire(r0,z0,0.2,pz0,5.0)
trajectoire(r0,z0,0.15,pz0,5.0)
trajectoire(r0,z0,0.1,pz0,5.0)
trajectoire(r0,z0,0.08,pz0,5.0)
trajectoire(r0,z0,0.05,pz0,5.0)
trajectoire(r0,z0,0.02,pz0,5.0)
axis([-1,1,-0.3,0.3])
xlabel('z')
ylabel('r')
grid()
figure(figsize=(8,8))
pz0=0.85
z0 = -0.7
pr0 = 0.1
tmax=5.0
r0=[0.0,0.01]
trajectoire(r0[0],z0,pr0,pz0,tmax)
trajectoire(r0[1],z0,pr0,pz0,tmax)
xlabel('z')
ylabel('r')
axis([-1,1,-0.3,0.3])
grid()
figure()
projection(r0[0],z0,pr0,pz0,tmax)
projection(r0[1],z0,pr0,pz0,tmax)
xlabel('x')
ylabel('y')
grid()
figure()
angle(r0[0],z0,pr0,pz0,tmax)
angle(r0[1],z0,pr0,pz0,tmax)
xlabel('z')
ylabel('theta')
grid()
show()